# 2.0.0 Migration Guide

The 2.0 release of the `google-cloud-iam` client is a significant upgrade based on a [next-gen code generator](https://github.com/googleapis/gapic-generator-python), and includes substantial interface changes. Existing code written for earlier versions of this library will likely require updates to use this version. This document describes the changes that have been made, and what you need to do to update your usage.

If you experience issues or have questions, please file an [issue](https://github.com/googleapis/python-iam/issues).

## Supported Python Versions

> **WARNING**: Breaking change

The 2.0.0 release requires Python 3.6+.


## Method Calls

> **WARNING**: Breaking change

Methods expect request objects. We provide a script that will convert most common use cases.

* Install the library

```py
python3 -m pip install google-cloud-iam[fixup]
```

* The script `fixup_credentials_v1_keywords.py` is shipped with the library. It expects an input directory (with the code to convert) and an empty destination directory.

```sh
$ fixup_credentials_v1_keywords.py --input-directory .samples/ --output-directory samples/
```

**Before:**
```py
from google.cloud import iam_credentials

client = iam_credentials.IAMCredentialsClient()

access_token = client.generate_access_token(
    name, scope, delegates = delegates, lifetime = lifetime
)
```


**After:**
```py
from google.cloud import iam_credentials

client = iam_credentials.IAMCredentialsClient()

access_token = client.generate_access_token(
    request = {'name': name, 'scope': scope, 'delegates': delegates, 'lifetime': lifetime}
)
```

### More Details

In `google-cloud-iam<2.0.0`, parameters required by the API were positional parameters and optional parameters were keyword parameters.

**Before:**
```py
    def generate_access_token(
        self,
        name,
        scope,
        delegates=None,
        lifetime=None,
        retry=google.api_core.gapic_v1.method.DEFAULT,
        timeout=google.api_core.gapic_v1.method.DEFAULT,
        metadata=None,
    ):
```

In the 2.0.0 release, all methods have a single positional parameter `request`. Method docstrings indicate whether a parameter is required or optional.

Some methods have additional keyword only parameters. The available parameters depend on the [`google.api.method_signature` annotation](https://github.com/googleapis/googleapis/blob/master/google/iam/credentials/v1/iamcredentials.proto#L49) specified by the API producer.


**After:**
```py
    def generate_access_token(
        self,
        request: common.GenerateAccessTokenRequest = None,
        *,
        name: str = None,
        scope: Sequence[str] = None,
        delegates: Sequence[str] = None,
        lifetime: duration.Duration = None,
        retry: retries.Retry = gapic_v1.method.DEFAULT,
        timeout: float = None,
        metadata: Sequence[Tuple[str, str]] = (),
    ) -> common.GenerateAccessTokenResponse:
```

> **NOTE:** The `request` parameter and flattened keyword parameters for the API are mutually exclusive.
> Passing both will result in an error.


Both of these calls are valid:

```py
access_token = client.generate_access_token(
    request={
        "name": name,
        "scope": scope,
        "delegates": delegates,
        "lifetime": lifetime
    }
)
```

```py
access_token = client.generate_access_token(
    name, scope, delegates = delegates, lifetime = lifetime
)
```

This call is invalid because it mixes `request` with a keyword argument `lifetime`. Executing this code will result in an error.

```py
access_token = client.generate_access_token(
    request={
        "name": name,
        "scope": scope,
        "delegates": delegates
    },
    lifetime = lifetime
)
```



## Types


> **WARNING**: Breaking change

The submodule `types` has been removed.

**Before:**
```py
from google.cloud import iam_credentials

request = iam_credentials.types.GenerateAccessTokenRequest(name="name")
```


**After:**
```py
from google.cloud import iam_credentials

request = iam_credentials.GenerateAccessTokenRequest(name="name")
```
